local _,N = ...
local AlertFrame, C_TalkingHead, C_PetBattles, C_Timer, CreateFrame, CloseGossip, CloseItemText, CloseQuest, CloseQuestChoice, GameTooltip, GetFriendshipReputation, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetGossipText, GetGreetingText, GetNumGossipActiveQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetNumQuestRewards, GetProgressText, GetQuestItemInfo, GetQuestText, GetRewardText, GetTime, GetTitleText, InCombatLockdown, IsInInstance, ITEM_TEXT_FROM, ItemRefTooltip, ItemTextGetCreator, ItemTextGetItem, ItemTextGetPage, ItemTextGetText, ItemTextHasNextPage, ItemTextNextPage, ItemTextPrevPage, ipairs, LibStub, QuestGetAutoAccept, QuestIsFromAreaTrigger, QuestIsFromAdventureMap, select, SelectGossipOption, tostring, tremove, UIParent, UnitGUID, UnitIsDead, UnitName, unpack
=     AlertFrame, C_TalkingHead, C_PetBattles, C_Timer, CreateFrame, CloseGossip, CloseItemText, CloseQuest, CloseQuestChoice, GameTooltip, GetFriendshipReputation, GetGossipActiveQuests, GetGossipAvailableQuests, GetGossipOptions, GetGossipText, GetGreetingText, GetNumGossipActiveQuests, GetNumGossipAvailableQuests, GetNumGossipOptions, GetNumQuestChoices, GetNumQuestRewards, GetProgressText, GetQuestItemInfo, GetQuestText, GetRewardText, GetTime, GetTitleText, InCombatLockdown, IsInInstance, ITEM_TEXT_FROM, ItemRefTooltip, ItemTextGetCreator, ItemTextGetItem, ItemTextGetPage, ItemTextGetText, ItemTextHasNextPage, ItemTextNextPage, ItemTextPrevPage, ipairs, LibStub, QuestGetAutoAccept, QuestIsFromAreaTrigger, QuestIsFromAdventureMap, select, SelectGossipOption, tostring, tremove, UIParent, UnitGUID, UnitIsDead, UnitName, unpack

local npcIds = LibStub("LibIDs").NPCS
local speed = 0.5


local ANIM_FREEZE = 3
local ANIM_SAY = 60
local ANIM_YELL = 64
local ANIM_ASK = 65
local ANIM_READ = 520


local function AnimFinished(self)
  if self.anim == ANIM_FREEZE or self.anim == ANIM_READ then return end
  if self.shouldLoop then
    self:SetAnimation(self.anim, 0)
  else
    self.anim = nil
  end
end

local function SetupAnimations(self, anim)
  if self.player or self.guid == UnitGUID("player") then
    anim = ANIM_READ
  elseif self.freeze then
    anim = ANIM_FREEZE
  end
  if self.anim == anim then return end
  self.anim = anim
  local should
  if self.static then
    should = N.Handler.core.chat.animate
  else
    should = N.Handler.core.conversation.animate
  end
  if anim == nil or self.freeze or should == "never" then
    self.shouldLoop = false
    self:StopAnimating()
    return
  end

  self:SetAnimation(anim, 0)
  self.shouldLoop = should == "loop"
end

local function queueItems()
  for i=1,GetNumQuestRewards() do
    GetQuestItemInfo("reward", i)
  end
  for i=1,GetNumQuestChoices() do
    GetQuestItemInfo("choice", i)
  end
end

local function ShowLink(self,_,link)
  GameTooltip:SetOwner(self, "ANCHOR_TOPLEFT", 10, 0)
  GameTooltip:SetHyperlink(link)
  GameTooltip:Show()
end

local frame = CreateFrame("Button", "Conversation", nil, "ConversationTemplate")
local chat = CreateFrame("Button", "ConversationChat", nil, "ConversationTemplate")

function frame:AnchorAlerts(conf, priority)
  priority = priority or self.anchorPriority or 1
  local shouldAnchor = conf.adjustAlerts
  if shouldAnchor then
    if self.alertAnchor then return end
    self:ClearAllPoints()
    self.alertAnchor = AlertFrame:AddJustAnchorFrameSubSystem(self.TopAnchor)
    local alertAnchor = self.alertAnchor
    alertAnchor.AdjustAnchors = function(_, relativeAlert)
      if not self:IsShown() then return relativeAlert end
      if relativeAlert == AlertFrame then
        self:SetPoint("BOTTOM", 0, 96)
      else
        self:SetPoint("BOTTOM", relativeAlert, "TOP", 0, 10)
      end
      return self.TopAnchor
    end
    AlertFrame:SetSubSustemAnchorPriority(alertAnchor, priority)
  else
    local anchor = self.alertAnchor
    if not anchor then return end
    self.alertAnchor = nil
    local systems = AlertFrame.alertFrameSubSystems
    for i,v in ipairs(systems) do
      if v == anchor then
        tremove(systems, i)
        return
      end
    end
    N.updateFrame(self, conf)
  end
end
chat.AnchorAlerts = frame.AnchorAlerts

frame:SetScript("OnLoad", function(self)
  self:RegisterEvent("LOADING_SCREEN_ENABLED")
  self:RegisterEvent("GOSSIP_SHOW")
  self:RegisterEvent("GOSSIP_CLOSED")
  self:RegisterEvent("QUEST_GREETING")
  self:RegisterEvent("QUEST_FINISHED")
  self:RegisterEvent("QUEST_DETAIL")
  self:RegisterEvent("QUEST_PROGRESS")
  self:RegisterEvent("QUEST_COMPLETE")
  self:RegisterEvent("QUEST_ITEM_UPDATE")

  self:RegisterEvent("ITEM_TEXT_BEGIN")
  self:RegisterEvent("ITEM_TEXT_TRANSLATION")
  self:RegisterEvent("ITEM_TEXT_READY")
  self:RegisterEvent("ITEM_TEXT_CLOSED")

  self:EnableKeyboard(true)

  self.NameFrame.Name:SetPoint("TOPLEFT", self.PortraitFrame.Portrait, "TOPRIGHT", 2, -19)
  self.PortraitFrame.Friendship:Show()

  local text = self.TextFrame
  text:SetScript("OnHyperlinkEnter", ShowLink)
  text:EnableMouse(true)
  text:SetScript("OnMouseDown", function(_, button)
    if button == "LeftButton" then
      self:Next()
      return true
    end
    return false
  end)

  for i=1,select("#", self:GetChildren()) do
    local child = select(i, self:GetChildren())
    for j=1,select("#", child:GetAnimationGroups()) do
      local animGroup = select(j, child:GetAnimationGroups())
      for k=1,select("#", animGroup:GetAnimations()) do
        local anim = select(k, animGroup:GetAnimations())
        anim:SetStartDelay(anim:GetStartDelay() * speed)
        anim:SetDuration(anim:GetDuration() * speed)
      end
    end
  end

  local model = self.MainFrame.Model
  model.SetupAnimations = SetupAnimations
  model:SetPortraitZoom(0.85)

  model:SetScript("OnAnimFinished", AnimFinished)

  self.TextFrame.Progress:SetScript("OnFinished", function()
    self:Next()
  end)
end)

chat:SetScript("OnLoad", function(self)
  self.static = true
  self:RegisterEvent("LOADING_SCREEN_ENABLED")
  self:RegisterEvent("CHAT_MSG_MONSTER_YELL")
  self:RegisterEvent("CHAT_MSG_MONSTER_SAY")
  self:RegisterEvent("CHAT_MSG_MONSTER_WHISPER")

  self.NameFrame.Name:SetPoint("TOPLEFT", self.PortraitFrame.Portrait, "TOPRIGHT", 2, -19)

  local text = self.TextFrame
  text:EnableMouse(true)
  text:SetScript("OnMouseDown", function(_, button)
    self:Close()
    return true
  end)

  for i=1,select("#", self:GetChildren()) do
    local child = select(i, self:GetChildren())
    for j=1,select("#", child:GetAnimationGroups()) do
      local animGroup = select(j, child:GetAnimationGroups())
      for k=1,select("#", animGroup:GetAnimations()) do
        local anim = select(k, animGroup:GetAnimations())
        anim:SetStartDelay(anim:GetStartDelay() * speed)
        anim:SetDuration(anim:GetDuration() * speed)
      end
    end
  end

  self.TextFrame.Continue:Hide()
  self.ChoicesFrame:Hide()
  self.TextFrame.ProgressLine:Hide()

  local model = self.MainFrame.Model
  model.SetupAnimations = SetupAnimations
  model:SetPortraitZoom(0.85)

  model:SetScript("OnAnimFinished", AnimFinished)

  self.anchorPriority = 10
end)

function frame:SetTop(anchor)
  self.TopAnchor:SetPoint("TOP", anchor or self)
end
chat.SetTop = frame.SetTop

local delay

local function Delay(funcName, ...)
  local wait
  if GetGossipOptions() or GetGossipActiveQuests() or GetGossipAvailableQuests() then
    wait = 0.1
  else wait = 0.3 end
  local root = ...
  local n = select("#", ...)
  local t = { ... }
  delay = funcName
  C_Timer.After(wait, function()
    if delay ~= funcName then return end
    root[funcName](unpack(t,1,n))
  end)
end

local function rewardText()
  local text = ""
  if N.Handler.core.conversation.completeRewards then
    text = N.getRewards(true, true) or ""
    if #text > 0 then text = "\r\r" .. text end
  end
  return text
end

function frame:RevealUI()
  if not UIParent:IsShown() or GameTooltip:GetParent() == self then
    UIParent:Show()
    ItemRefTooltip:SetParent(UIParent)
    GameTooltip:SetParent(UIParent)
    ItemRefTooltip:SetFrameStrata("TOOLTIP")
    GameTooltip:SetFrameStrata("TOOLTIP")
    ItemRefTooltip:Raise()
    GameTooltip:Raise()
  end
end

frame:SetScript("OnEvent", function(self, event, ...)
  N.Handler:Output(event)
  if event == "LOADING_SCREEN_ENABLED" then
    self:Reset()
    self:CloseImmediately()
    return
  elseif event == "GOSSIP_SHOW" then
    if GetNumGossipAvailableQuests() == 0 and GetNumGossipActiveQuests() == 0 and GetNumGossipOptions() == 1 then
      if select(2, GetGossipOptions()) ~= "gossip" then
        SelectGossipOption(1)
        return
      end
    end
    Delay("PlayCurrent", self, UnitName("npc"), GetGossipText())
  elseif event == "GOSSIP_CLOSED" then
    Delay("Close", self, true)
  elseif event == "QUEST_GREETING" then
    Delay("PlayCurrent", self, UnitName("questnpc") or UnitName("npc"), GetGreetingText())
  elseif event == "QUEST_FINISHED" then
    Delay("Close", self)
  elseif event == "QUEST_DETAIL" then
    if QuestIsFromAdventureMap() or (QuestGetAutoAccept() and QuestIsFromAreaTrigger()) or (... ~= nil and ... ~= 0) or GetQuestText() == "" then
      return
    end
    queueItems()
    local conf = N.Handler.core.conversation
    Delay("PlayCurrent", self, GetTitleText(), GetQuestText() .. N.plusObjective(event, conf.rewards, false))
  elseif event == "QUEST_PROGRESS" then
    queueItems()
    local conf = N.Handler.core.conversation
    Delay("PlayCurrent", self, GetTitleText(), GetProgressText() .. N.plusObjective(event, false, conf.finished))
  elseif event == "QUEST_COMPLETE" then
    Delay("PlayCurrent", self, GetTitleText(), GetRewardText() .. rewardText())
  elseif event == "QUEST_ITEM_UPDATE" then
    if self.lastEvent == "QUEST_COMPLETE" then
      self:SetText(GetRewardText() .. rewardText())
    end
    return
  elseif not N.Handler.core.conversation.read then
    return
  elseif event == "ITEM_TEXT_TRANSLATION" then
    if not self:IsShown() then CloseItemText() end
  elseif event == "ITEM_TEXT_READY" then
    local text = ItemTextGetText()
    local creator = ItemTextGetCreator()
    if creator then
      text = text .. "\r" .. ITEM_TEXT_FROM .. "\r" .. creator
    end
    Delay("PlayCurrent", self, ItemTextGetItem(), text)
  elseif event == "ITEM_TEXT_CLOSED" then
    Delay("Close", self)
  end
  self.lastEvent = event
end)

local lastTime
chat:SetScript("OnEvent", function(self, event, message, sender, ...)
  N.Handler:Output(event)
  local conf = N.Handler.core.chat
  if event == "LOADING_SCREEN_ENABLED" then
    self:Reset()
    self:CloseImmediately()
    return
  elseif C_TalkingHead.GetCurrentLineInfo() or C_PetBattles.IsInBattle() then
    return
  elseif (event == "CHAT_MSG_MONSTER_YELL" and not conf.yell)
      or (event == "CHAT_MSG_MONSTER_SAY" and not conf.say)
      or (event == "CHAT_MSG_MONSTER_WHISPER" and not conf.whisper) then
    return
  elseif conf[InCombatLockdown() and "incombat" or "outofcombat"] then
    local instance, instanceType = IsInInstance()
    if not instance then if not conf.none then return end
    elseif instanceType == "raid" then if not conf.raid then return end
    else if not conf.dungeon then return end
    end
    lastTime = GetTime()
    local l = lastTime
    self:PlayCurrent(sender, message)
    C_Timer.After(N.Handler.core.chat.slideSpeed, function()
      if lastTime == l then self:Close() end
    end)
  end
end)

frame:SetScript("OnShow", function(self)
  if N.Handler.core.conversation.conceal then
    ItemRefTooltip:SetParent(self)
    GameTooltip:SetParent(self)
    ItemRefTooltip:SetFrameStrata("TOOLTIP")
    GameTooltip:SetFrameStrata("TOOLTIP")
    UIParent:Hide()
  end
  self.TopAnchor:Show()
  AlertFrame:UpdateAnchors()
end)
chat:SetScript("OnShow", function(self)
  AlertFrame:UpdateAnchors()
end)
frame:SetScript("OnHide", function(self)
  N.Handler:HideBlizzard()
  self:RevealUI()
  self.TopAnchor:Hide()
  AlertFrame:UpdateAnchors()
end)
chat:SetScript("OnHide", function(self)
  AlertFrame:UpdateAnchors()
  self.MainFrame.Model:ClearModel()
end)

local function endInteraction(partial)
  N.Handler:Output("Closing ",partial)
  if not partial then
    CloseGossip()
    CloseQuestChoice()
    if N.Handler.core.conversation.read then CloseItemText() end
  end
  CloseQuest()
end

frame:SetScript("OnKeyDown", function(self, key)
  key = N.keybindmods() .. key
  N.Handler:Output("Key down: ",key)
  local conf = N.Handler.core.conversation
  self:SetPropagateKeyboardInput(false)
  if key == conf.next then
    self:Next()
    return
  elseif key == conf.back then
    if self.lastEvent == "ITEM_TEXT_READY" and ItemTextGetPage() > 1 then
      ItemTextPrevPage()
    else
      self:FadeText(self.full)
    end
    return
  elseif key == conf.close then
    endInteraction()
  elseif key == conf.auto then
    local child = self.ChoicesFrame.auto
    if child then
      child:GetScript("OnEnter")(child)
      child:Click()
    else
      self:Next()
    end
    self:SetPropagateKeyboardInput(false)
    return
  elseif key == conf.reveal then
    N.Handler:ShowBlizzard()
    self:RevealUI()
  else
    local keyNumber
    for i=1,12 do
      if conf[tostring(i)] == key then
        keyNumber = i
        break
      end
    end
    local sum = self.ChoicesFrame.sum
    if keyNumber and keyNumber <= sum then
      local child = select(sum + 1 - keyNumber, self.ChoicesFrame:GetChildren())
      if child.choose then
        child:GetScript("OnEnter")(child)
        child:Click()
      end
      self:SetPropagateKeyboardInput(false)
      return
    end
  end

  self:SetPropagateKeyboardInput(true)
  return
end)

function frame:CloseImmediately()
  self:Hide()
  endInteraction()
end
chat.CloseImmediately = frame.CloseImmediately

function frame:FadeinFrames()
  self.MainFrame.TalkingHeadsInAnim:Play()
  C_Timer.After(0.5 * speed, function()
    self.NameFrame.Fadein:Play()
  end)
  C_Timer.After(0.75 * speed, function()
    self.TextFrame.Fadein:Play()
  end)
  self.BackgroundFrame.Fadein:Play()
  self.PortraitFrame.Fadein:Play()
end
chat.FadeinFrames = frame.FadeinFrames

function frame:FadeoutFrames()
  if concealed then
    concealed = nil
    UIParent:Show()
  end
  self.ChoicesFrame:Close()
  self.MainFrame.Close:Play()
  self.NameFrame.Close:Play()
  self.TextFrame.Close:Play()
  self.BackgroundFrame.Close:Play()
  self.PortraitFrame.Close:Play()
  if self.TextFrame.Continue:GetAlpha() ~= 0 then
    self.TextFrame.FadeoutContinue:Play()
  end
  if self.PortraitFrame.Friendship:GetAlpha() ~= 0 then
    self.PortraitFrame.FadeoutFriendship:Play()
  end
  self.TextFrame.Progress:Stop()
end
function chat:FadeoutFrames()
  self.MainFrame.Close:Play()
  self.NameFrame.Close:Play()
  self.TextFrame.Close:Play()
  self.BackgroundFrame.Close:Play()
  self.PortraitFrame.Close:Play()
  self.TextFrame.Progress:Stop()
end

function frame:SetText(text)
  if not text or #text == 0 then
    self.TextFrame.Text:SetText()
  else
    text = text:trim()
    local line = text:find("\n")
    if line then
      self.remaining = text:sub(line)
      --text = text:sub(1,line):trim()
      text = text:sub(1,line)
    else
      self.remaining = ""
    end
    local t = self.TextFrame.Text
    t:SetText(text:trim())
    if text:sub(1,1) ~= "|" then
      local anim
      if text:find("?") then
        anim = ANIM_ASK
      elseif text:find("!") then
        anim = ANIM_YELL
      else
        anim = ANIM_SAY
      end
      self.MainFrame.Model:SetupAnimations(anim)
    end
    local trail
    for _,delim in ipairs{ "\n\r", "\.?!", ",:"} do
      local sub = t:GetText()
      while t:IsTruncated() do
        sub, trail = select(3, sub:find(
          "(.+[" .. delim .. "])(.+[^\s" .. delim .. "].+)"
        ))
        if not sub then break end
        --sub = sub:trim()
        if #sub == 0 then break end
        self.remaining = trail .. self.remaining
        t:SetText(sub)
      end
    end
  end
  if (self.lastEvent ~= "ITEM_TEXT_READY" or not ItemTextHasNextPage()) and (not self.remaining or #self.remaining == 0) then
    if self.TextFrame.Continue:GetAlpha() ~= 0 then
      self.TextFrame.FadeoutContinue:Play()
    end

    self.chill = true
    if not self.static then
      C_Timer.After(0.5 * speed, function()
        self.ChoicesFrame:Refresh(self.lastEvent)
        self.chill = nil
      end)
    end
  else
    if self.remaining:sub(1,2) == "|c" then
      self.TextFrame.Continue:SetText(self.remaining:sub(1,10) .. "...|r")
    else
      self.TextFrame.Continue:SetText("...")
    end
    if self.TextFrame.Continue:GetAlpha() == 0 then
      self.TextFrame.FadeinContinue:Play()
    end
    if not self.static and N.Handler.core.conversation.slideshow then
      self.TextFrame.Progress:Restart()
    end
    self.ChoicesFrame:Close()
  end

  if self.remaining then self.remaining = self.remaining:trim() end
  self.chill2 = nil
end
chat.SetText = frame.SetText

function frame:FadeText(text)
  self.TextFrame.Fadeout:Play()
  C_Timer.After(0.25 * speed, function()
    self:SetText(text)
  end)
  C_Timer.After(0.5 * speed, function()
    self.TextFrame.Fadein:Play()
  end)
end
chat.FadeText = frame.FadeText
function chat:SetText(text)
  if not text or #text == 0 then
    self.TextFrame.Text:SetText()
    return
  end
  self.TextFrame.Text:SetText(text:trim())
  local anim
  if text:find("?") then
    anim = ANIM_ASK
  elseif text:find("!") then
    anim = ANIM_YELL
  else
    anim = ANIM_SAY
  end
  self.MainFrame.Model:SetupAnimations(anim)
end
chat.SetText = frame.SetText

function frame:FadeText(text)
  self.TextFrame.Fadeout:Play()
  C_Timer.After(0.25 * speed, function()
    self:SetText(text)
  end)
  C_Timer.After(0.5 * speed, function()
    self.TextFrame.Fadein:Play()
  end)
end

function frame:Reset(text, name)
  self:StopAnimating()
  self.BackgroundFrame.TextBackground:SetAlpha(0.01)
  self.NameFrame.Name:SetAlpha(0.01)
  self.TextFrame.Text:SetAlpha(0.01)
  self.MainFrame.Sheen:SetAlpha(0.01)
  self.MainFrame.TextSheen:SetAlpha(0.01)

  self.MainFrame.Model:SetAlpha(0.01)
  self.MainFrame.Model.PortraitBg:SetAlpha(0.01)
  self.PortraitFrame.Portrait:SetAlpha(0.01)
  self.MainFrame.Overlay.Glow_LeftBar:SetAlpha(0.01)
  self.MainFrame.Overlay.Glow_RightBar:SetAlpha(0.01)
  self.MainFrame.CloseButton:SetAlpha(0.01)

  self.MainFrame:SetAlpha(1)
  self.NameFrame.Name:SetText(name)
  self:SetText(text)
end
chat.Reset = frame.Reset

local htmlTag = "(?i)<\\s*(p|[h[1-3]|align|img|a)(\\s+[^>]*>|/?>)"
function frame:PlayCurrent(name, text, new)
  if text:find("<") then
    text = text:gsub("<\s*/?[bB][rR][^>]*>", "\r")
    for _,el in pairs{
      "[hH][tT][mM][lL",
      "[bB][oO][dD][yY]",
      "[hH][eE][aA][dD]",
      "[pP]",
      "[hH]%d",
      "[aA][lL][iI][gG][nN]",
      "[iI][mM][gG]",
      "[aA]",
      "[hH][rR]"
    } do
      text = text:gsub("<\s*/?" .. el .. " [^>]*/?>",""):gsub("<\s*/?" .. el .. "/?>","")
    end
  end
  text = text:trim()
  self.chill2 = true
  if not new and (not self:IsShown() or self.MainFrame.CloseButton:GetAlpha() ~= 1) then new = true end
  local model = self.MainFrame.Model

  local textFormatted = format(text)
  self.full = textFormatted
  if text then
    self:Show()

    if self.static then
      if model.name ~= name then
        if new == nil then new = true end
        model.name = name
        model:ClearModel()
      end
      local id = npcIds[name]
      N.Handler:Output("ID for ",name," is ",id or "not found")
      if id then
        model:SetCreature(id)
        model:SetupAnimations()
      end
    else
      local unit = "questnpc"
      local guid = UnitGUID(unit)
      if not guid then
        unit = "npc"
        guid = UnitGUID(unit)
      end
      model.freeze = UnitIsDead(unit)

      if model.guid ~= guid then
        model.anim = nil
        if new == nil then new = true end
        model.guid = guid
        local prefId = guid:sub(1,4)
        if guid == UnitGUID("player") or prefId == "Item" or prefId == "Game" then
          model.player = true
          model:SetUnit("player")
        else
          model.player = false
          model:SetUnit(unit)
        end
        model:SetupAnimations()
      end
    end

    if new then
      self.ChoicesFrame:Hide()
      self:Reset(textFormatted, name)
      self:FadeinFrames()
    else
      if name ~= self.NameFrame.Name:GetText() then
        -- Fade out the old name and fade in the new name
        self.NameFrame.Fadeout:Play()
        C_Timer.After(0.25 * speed, function()
          self.NameFrame.Name:SetText(name)
        end)
        C_Timer.After(0.5 * speed, function()
          self.NameFrame.Fadein:Play()
        end)

      end

      self:FadeText(textFormatted)
    end
  end
  if self.UpdateFriendship then self:UpdateFriendship() end
end
chat.PlayCurrent = frame.PlayCurrent

function frame:Next()
  if self.MainFrame.CloseButton:GetAlpha() ~= 1 then return end
  if self.remaining and #(self.remaining) > 0 then
    self:FadeText(self.remaining)
  elseif self.lastEvent == "ITEM_TEXT_READY" and ItemTextHasNextPage() then
    ItemTextNextPage()
  elseif self.ChoicesFrame.sum == 0 and not self.chill and not self.chill2 then
    endInteraction()
  end
end

function frame:Close(partial)
  self.MainFrame.Model:StopAnimating()
  self:FadeoutFrames()
  endInteraction(partial)
end
function chat:Close()
  self.MainFrame.Model:StopAnimating()
  self:FadeoutFrames()
end

function frame:UpdateFriendship()
  local conf = N.Handler.core.conversation
  local statusBar = self.PortraitFrame.Friendship
  local id, rep, maxRep, name, text, texture, reaction, threshold, nextThreshold = GetFriendshipReputation()
  if id and id > 0 and conf.friendship then
    -- if max rank, make it look like a full bar
    if not nextThreshold then
      threshold, nextThreshold, rep = 0, 1, 1
    end
    if texture then
      statusBar.icon:SetTexture(texture)
    else
      statusBar.icon:SetTexture("Interface\\Common\\friendship-heart")
    end
    statusBar:SetMinMaxValues(threshold, nextThreshold)
    local color = conf.friendshipColor
    statusBar:SetStatusBarColor(color.r, color.g, color.b, color.a)
    statusBar:SetValue(rep)
    statusBar:EnableMouse(true)
    self.PortraitFrame.FadeinFriendship:Play()
  else
    if statusBar:GetAlpha() ~= 0 then
      self.PortraitFrame.FadeoutFriendship:Play()
    end
    statusBar:EnableMouse(false)
  end
end

frame:GetScript("OnLoad")(frame)
chat:GetScript("OnLoad")(chat)